package com.astro.bibliotheca.network;

import io.netty.buffer.ByteBuf;
import net.minecraft.item.ItemStack;
import net.minecraft.nbt.NBTTagCompound;
import net.minecraft.network.PacketBuffer;
import net.minecraft.util.math.BlockPos;
import net.minecraftforge.fml.common.network.simpleimpl.IMessage;
import net.minecraftforge.fml.common.network.simpleimpl.IMessageHandler;
import net.minecraftforge.fml.common.network.simpleimpl.MessageContext;
import org.apache.commons.lang3.tuple.Pair;

import java.io.IOException;
import java.io.Serializable;
import java.lang.reflect.Field;
import java.lang.reflect.Modifier;
import java.util.HashMap;

public abstract class Message<REQ extends Message<REQ>> implements Serializable,IMessage,IMessageHandler<REQ,IMessage> {
    private static final HashMap<Class, Pair<Reader, Writer>> handlers = new HashMap<>();

    static {
        map(byte.class, PacketBuffer::readByte, (Writer<Byte>) PacketBuffer::writeByte);
        map(short.class, PacketBuffer::readShort, (Writer<Short>) PacketBuffer::writeShort);
        map(int.class, PacketBuffer::func_150792_a, PacketBuffer::func_150787_b);
        map(long.class, PacketBuffer::func_179260_f, PacketBuffer::func_179254_b);
        map(float.class, PacketBuffer::readFloat, PacketBuffer::writeFloat);
        map(double.class, PacketBuffer::readDouble, PacketBuffer::writeDouble);
        map(boolean.class, PacketBuffer::readBoolean, PacketBuffer::writeBoolean);
        map(char.class, PacketBuffer::readChar, (Writer<Character>) PacketBuffer::writeChar);
        map(String.class, Message::readString, PacketBuffer::func_180714_a);
        map(NBTTagCompound.class, PacketBuffer::func_150793_b, PacketBuffer::func_150786_a);
        map(ItemStack.class, PacketBuffer::func_150791_c, PacketBuffer::func_150788_a);
        map(BlockPos.class, PacketBuffer::func_179259_c, PacketBuffer::func_179255_a);
    }

    private static String readString(PacketBuffer buf) {
       return buf.func_150789_c(32767);
    }

    @Override
    public IMessage onMessage(REQ message, MessageContext ctx) {
        return message.handleMessage(ctx);
    }

    protected abstract IMessage handleMessage(MessageContext context);

    @Override
    public final void fromBytes(ByteBuf buf) {
        try {
            PacketBuffer packetBuffer = new PacketBuffer(buf);
            Class<?> clazz = getClass();
            for (Field f : clazz.getDeclaredFields()) {
                Class<?> type = f.getType();
                if (acceptField(f, type))
                    readField(f, type, packetBuffer);
            }
        } catch (Exception e) {
            throw new RuntimeException("Error at reading packet " + this, e);
        }
    }

    private void writeField(Field f, Class clazz, PacketBuffer buf) throws IllegalArgumentException, IllegalAccessException {
        Pair<Reader, Writer> handler = getHandler(clazz);
        handler.getRight().write(buf, f.get(this));
    }

    private void readField(Field f, Class clazz, PacketBuffer buf) throws IllegalArgumentException, IllegalAccessException, IOException {
        Pair<Reader, Writer> handler = getHandler(clazz);
        f.set(this, handler.getLeft().read(buf));
    }

    private static Pair<Reader, Writer> getHandler(Class<?> clazz) {
        Pair<Reader, Writer> pair = handlers.get(clazz);
        if (pair == null)
            throw new RuntimeException("No R/W handler for  " + clazz);
        return pair;
    }

    private static boolean acceptField(Field f, Class<?> type) {
        int mods = f.getModifiers();
        return !Modifier.isFinal(mods) && !Modifier.isStatic(mods) && !Modifier.isTransient(mods) && handlers.containsKey(type);

    }

    @Override
    public final void toBytes(ByteBuf buf) {
        try {
            PacketBuffer packetBuffer = new PacketBuffer(buf);
            Class<?> clazz = getClass();
            for (Field f : clazz.getDeclaredFields()) {
                Class<?> type = f.getType();
                if (acceptField(f, type))
                    writeField(f, type, packetBuffer);
            }
        } catch (Exception e) {
            throw new RuntimeException("Error at writing packet " + this, e);
        }
    }

    private static <T> void map(Class<T> type, Reader<T> reader, Writer<T> writer) {
        handlers.put(type, Pair.of(reader, writer));
    }

    public interface Writer<T> {
        void write(PacketBuffer buf, T t);
    }

    public interface Reader<T> {
        T read(PacketBuffer buf) throws IOException;
    }
}
